////////////////////////////////////////////////////////////////////////////////
// CONFIDENTIAL and PROPRIETARY software of Magewell Electronics Co., Ltd.
// Copyright (c) 2011-2018 Magewell Electronics Co., Ltd. (Nanjing)
// All rights reserved.
// This copyright notice MUST be reproduced on all authorized copies.
////////////////////////////////////////////////////////////////////////////////
#include "ospi.h"

#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/kthread.h>
#include <linux/err.h>
#include <linux/pagemap.h>
#include <linux/dma-mapping.h>
#include <linux/cache.h>
#include <linux/spinlock.h>
#include <linux/vmalloc.h>
#include <linux/version.h>
#include <linux/printk.h>
#include <linux/freezer.h>
#include <linux/pci.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
#include <linux/sched/signal.h>
#include <linux/sched/debug.h>
#endif

unsigned int debug_level = 0;

struct kmem_cache *g_spin_cache = NULL;
struct kmem_cache *g_mutex_cache = NULL;
struct kmem_cache *g_event_cache = NULL;

/* init */
int os_init(void)
{
    int ret;
    size_t size;

    /* kmem_cache_create require size >= BYTES_PER_WORD && size <= KMALLOC_MAX_SIZE */
    size = sizeof(spinlock_t) < sizeof(void *) ? sizeof(void *) : sizeof(spinlock_t);
    g_spin_cache = kmem_cache_create("mw-spin", size,
                                     0, 0, NULL);
    if (g_spin_cache == NULL)
        return OS_RETURN_NOMEM;

    g_mutex_cache = kmem_cache_create("mw-mutex", sizeof(struct mutex),
                                     0, 0, NULL);
    if (g_mutex_cache == NULL) {
        ret = OS_RETURN_NOMEM;
        goto mutex_err;
    }

    g_event_cache = kmem_cache_create("mw-event", sizeof(struct _os_event_t),
                                     0, 0, NULL);
    if (g_event_cache == NULL) {
        ret = OS_RETURN_NOMEM;
        goto event_err;
    }

    return OS_RETURN_SUCCESS;

event_err:
    kmem_cache_destroy(g_mutex_cache);
mutex_err:
    kmem_cache_destroy(g_spin_cache);
    return ret;
}

void os_deinit(void)
{
    if (g_spin_cache != NULL)
        kmem_cache_destroy(g_spin_cache);
    if (g_mutex_cache != NULL)
        kmem_cache_destroy(g_mutex_cache);
    if (g_event_cache != NULL)
        kmem_cache_destroy(g_event_cache);

    g_spin_cache = NULL;
    g_mutex_cache = NULL;
    g_event_cache = NULL;
}


/* print */
int __os_printf(const char *s, ...)
{
    int retval;
    va_list ap;

    va_start(ap, s);
    retval = vprintk(s, ap);
    va_end(ap);

    return retval;
}


/* time */
os_clock_kick_t os_get_clock_kick(void)
{
    return jiffies;
}

uint64_t os_clock_kick_to_msecs(os_clock_kick_t kick)
{
    return jiffies_to_msecs(kick);
}

os_clock_kick_t os_msecs_to_clock_kick(uint64_t msecs)
{
    return msecs_to_jiffies(msecs);
}

/* time_after(a,b) returns true if the time a is after time b. */
int os_clock_kick_after(uint64_t a, uint64_t b)
{
    return time_after((unsigned long)a, (unsigned long)b);
}

int os_clock_kick_after_eq(uint64_t a, uint64_t b)
{
    return time_after_eq((unsigned long)a, (unsigned long)b);
}


/* timer */
struct _os_timer {
    struct timer_list   k_timer;
    timeout_func_t      func;
    void *              param;
};

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
static void __os_timer_callback_wrapper(unsigned long param)
{
    os_timer timer = (os_timer)param;

    if (timer && timer->func)
        timer->func(timer->param);
}
#else
static void __os_timer_callback_wrapper(struct timer_list *t)
{
    os_timer timer = from_timer(timer, t, k_timer);

    if (timer && timer->func)
        timer->func(timer->param);
}
#endif

os_timer os_timer_alloc(timeout_func_t timeout_func, void *param)
{
    os_timer timer;

    timer = kzalloc(sizeof(*timer), GFP_KERNEL);
    if (timer == NULL)
        return NULL;

    timer->func = timeout_func;
    timer->param = param;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
    setup_timer(&timer->k_timer, __os_timer_callback_wrapper, (unsigned long)timer);
#else
    timer_setup(&timer->k_timer, __os_timer_callback_wrapper, 0);
#endif

    return timer;
}

void os_timer_free(os_timer timer)
{
    del_timer(&timer->k_timer);
    kfree(timer);
}

void os_timer_schedule_clock_kick(os_timer timer, os_clock_kick_t kick)
{
    mod_timer(&timer->k_timer, kick);
}

void os_timer_schedule_relative(os_timer timer, uint32_t timeout_in_ms)
{
    os_clock_kick_t _time = msecs_to_jiffies(timeout_in_ms);

    mod_timer(&timer->k_timer, jiffies+_time);
}

void os_timer_cancel(os_timer timer)
{
    del_timer(&timer->k_timer);
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 0, 0))
void atomic_or(int i, atomic_t *v)
{
    int old;
    int new;

    do {
        old = atomic_read(v);
        new = old | i;
    } while (atomic_cmpxchg(v, old, new) != old);
}
#endif

int os_atomic_read(const os_atomic_t *v)
{
    return atomic_read(v);
}

void os_atomic_set(os_atomic_t *v, int i)
{
    atomic_set(v, i);
}

int os_atomic_inc(os_atomic_t *v)
{
    return atomic_inc_return(v);
}

int os_atomic_dec(os_atomic_t *v)
{
    return atomic_dec_return(v);
}

int os_atomic_add(os_atomic_t *v, int i)
{
    return atomic_add_return(i, v);
}

int os_atomic_sub(os_atomic_t *v, int i)
{
    return atomic_sub_return(i, v);
}

void os_atomic_or(os_atomic_t *v, int i)
{
    atomic_or(i, v);
}

int os_atomic_xchg(os_atomic_t *v, int i)
{
    return atomic_xchg(v, i);
}

int os_atomic_cmpxchg(os_atomic_t *v, int oldv, int newv)
{
    return atomic_cmpxchg(v, oldv, newv);
}

void os_clear_bit(int nr, volatile unsigned long *addr)
{
    clear_bit(nr, addr);
}

void os_set_bit(int nr, volatile unsigned long *addr)
{
    set_bit(nr, addr);
}

int os_test_bit(int nr, const volatile unsigned long *addr)
{
    return test_bit(nr, addr);
}

int os_test_and_set_bit(int nr, volatile unsigned long *addr)
{
    return test_and_set_bit(nr, addr);
}

int os_test_and_clear_bit(int nr, volatile unsigned long *addr)
{
    return test_and_clear_bit(nr, addr);
}


/* spin lock */
os_spinlock_t os_spin_lock_alloc(void)
{
    os_spinlock_t lock;

    lock = kmem_cache_alloc(g_spin_cache, GFP_KERNEL);
    if (lock == NULL)
        return NULL;

    spin_lock_init((spinlock_t *)lock);

    return lock;
}

void os_spin_lock_free(os_spinlock_t lock)
{
    kmem_cache_free(g_spin_cache, lock);
}

void os_spin_lock(os_spinlock_t lock)
{
    spin_lock(lock);
}

void os_spin_unlock(os_spinlock_t lock)
{
    spin_unlock(lock);
}

void os_spin_lock_bh(os_spinlock_t lock)
{
    spin_lock_bh(lock);
}

void os_spin_unlock_bh(os_spinlock_t lock)
{
    spin_unlock_bh(lock);
}

int os_spin_try_lock(os_spinlock_t lock)
{
    return spin_trylock(lock);
}

/* mutex */
os_mutex_t os_mutex_alloc(void)
{
    os_mutex_t lock;

    lock = kmem_cache_alloc(g_mutex_cache, GFP_KERNEL);
    if (lock == NULL)
        return NULL;

    mutex_init(lock);

    return lock;
}

void os_mutex_free(os_mutex_t lock)
{
    kmem_cache_free(g_mutex_cache, lock);
}

void os_mutex_lock(os_mutex_t lock)
{
    mutex_lock(lock);
}

void os_mutex_unlock(os_mutex_t lock)
{
    mutex_unlock(lock);
}

int os_mutex_try_lock(os_mutex_t lock)
{
    return mutex_trylock(lock);
}

/* event */
static inline long
#ifdef CONFIG_PREEMPT_RT_FULL
do_wait_and_clear(struct rt_completion *x, long timeout, int state)
#else
do_wait_and_clear(struct completion *x, long timeout, int state)
#endif
{
    if (!x->done) {
        DECLARE_WAITQUEUE(wait, current);

        wait.flags |= WQ_FLAG_EXCLUSIVE;
        __add_wait_queue(&x->wait, &wait);
        do {
            if (signal_pending_state(state, current)) {
                timeout = -ERESTARTSYS;
                break;
            }
            __set_current_state(state);
            spin_unlock_irq(&x->wait.lock);
            timeout = schedule_timeout(timeout);
            spin_lock_irq(&x->wait.lock);
        } while (!x->done && timeout);
        __remove_wait_queue(&x->wait, &wait);
        if (!x->done)
            return timeout;
    }
    x->done = 0;
    return timeout ?: 1;
}

static long __sched
#ifdef CONFIG_PREEMPT_RT_FULL
wait_and_clear(struct rt_completion *x, long timeout, int state)
#else
wait_and_clear(struct completion *x, long timeout, int state)
#endif
{
    might_sleep();

    spin_lock_irq(&x->wait.lock);
    timeout = do_wait_and_clear(x, timeout, state);
    spin_unlock_irq(&x->wait.lock);
    return timeout;
}

os_event_t os_event_alloc(void)
{
    os_event_t event;

    event = kmem_cache_alloc(g_event_cache, GFP_KERNEL);
    if (event == NULL)
        return NULL;

#ifdef CONFIG_PREEMPT_RT_FULL
    event->done.done = 0;
    init_waitqueue_head(&event->done.wait);
#else
    init_completion(&event->done);
#endif

    os_atomic_set(&event->refcount, 1);

    return event;
}

void os_event_release(os_event_t event)
{
    if (os_atomic_dec(&event->refcount) == 0) {
        kmem_cache_free(g_event_cache, event);
    }
}

void os_event_set(os_event_t event)
{
#ifdef CONFIG_PREEMPT_RT_FULL
    unsigned long flags;

    spin_lock_irqsave(&event->done.wait.lock, flags);
    event->done.done++;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,4,0)
    __wake_up_locked(&event->done.wait, TASK_NORMAL, 1);
#else
    __wake_up_locked(&event->done.wait, TASK_NORMAL);
#endif
    spin_unlock_irqrestore(&event->done.wait.lock, flags);
#else /* CONFIG_PREEMPT_RT_FULL */
    complete(&event->done);
#endif /* CONFIG_PREEMPT_RT_FULL */
}

void os_event_clear(os_event_t event)
{
#ifdef CONFIG_PREEMPT_RT_FULL
    event->done.done = 0;
#else

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 13, 0))
#  define reinit_completion(x) INIT_COMPLETION(*(x))
#endif
    reinit_completion(&event->done);
#endif
}

bool os_event_is_set(os_event_t event)
{
#ifdef CONFIG_PREEMPT_RT_FULL
    unsigned long flags;
    int ret = 1;

    spin_lock_irqsave(&event->done.wait.lock, flags);
    if (!event->done.done)
        ret = 0;
    spin_unlock_irqrestore(&event->done.wait.lock, flags);
    return ret;
#else
    return completion_done(&event->done);
#endif
}

bool os_event_try_wait(os_event_t event)
{
    unsigned long flags;
    int ret = 1;

    spin_lock_irqsave(&event->done.wait.lock, flags);
    if (!event->done.done)
        ret = 0;
    else
        event->done.done = 0;
    spin_unlock_irqrestore(&event->done.wait.lock, flags);

    return ret;
}

/* @timeout: -1: wait for ever; 0: no wait; >0: wait for @timeout ms */
/* ret 0: timeout, <0 err, >0 event occur */
int32_t os_event_wait(os_event_t event, int32_t timeout)
{
    unsigned long expire;
    long ret = 0;

    if (timeout < 0)
        expire = MAX_SCHEDULE_TIMEOUT;
    else
        expire = msecs_to_jiffies(timeout);

    ret = wait_and_clear(&event->done, expire, TASK_KILLABLE);

    if (ret > INT_MAX)
        ret = 1;

    return ret;
}

/* @timeout: -1: wait for ever; 0: no wait; >0: wait for @timeout ms */
/* @ret: 0 timeout; >0 event occur */
int32_t os_event_wait_for_multiple(os_event_t events[], int num_events,
                                   int32_t timeout/* ms */)
{
    unsigned long expire;
    int i;
    bool got_event;
    int32_t ret = 0;

    if (timeout < 0)
        expire = INT_MAX; // MAX_SCHEDULE_TIMEOUT
    else
        expire = msecs_to_jiffies(timeout);

    for (i = 0; i < num_events; i++) {
        init_waitqueue_entry(&events[i]->waitq, current);
        /* add ourselves into wait queue */
        add_wait_queue(&events[i]->done.wait, &events[i]->waitq);
    }

    got_event = false;
    for (;;) {
        set_current_state(TASK_KILLABLE);

        for (i = 0; i < num_events; i++) {
            if (os_event_is_set(events[i])) {
                got_event = true;
                break;
            }
        }
        if (got_event) {
            ret = 1;
            break;
        }
        expire = schedule_timeout(expire);
        if (!expire) {
            ret = 0;
            break;
        }
    }

    __set_current_state(TASK_RUNNING);
    for (i = 0; i < num_events; i++) {
        remove_wait_queue(&events[i]->done.wait, &events[i]->waitq);
    }

    return ret;
}


/* thread */
struct _os_thread_t {
    struct task_struct          *kthread;
    os_event_t                  wait_exit;
};

os_thread_t os_thread_create_and_run(int (*threadfn)(void *data),
                                     void *data,
                                     const char *name)
{
    os_thread_t thread = (os_thread_t)os_zalloc(sizeof(*thread));

    if (thread == NULL)
        return NULL;

    thread->wait_exit = os_event_alloc();
    if (thread->wait_exit == NULL) {
        os_free(thread);
        return NULL;
    }

    thread->kthread = kthread_run(threadfn, data, name);
    if (IS_ERR_OR_NULL(thread->kthread)) {
        os_event_release(thread->wait_exit);
        os_free(thread);
        return NULL;
    }

    return thread;
}

void os_thread_terminate_self(os_thread_t thread)
{
    os_event_set(thread->wait_exit);
}

void os_thread_wait_and_free(os_thread_t thread)
{
    os_event_wait(thread->wait_exit, -1);
    os_event_release(thread->wait_exit);
    os_free(thread);
}


/* mem io */
uint32_t pci_read_reg32(volatile void *addr, os_pci_dev_t pci_dev) {
    return readl(addr);
}

void pci_write_reg32(volatile void *addr, uint32_t val, os_pci_dev_t pci_dev) {
    writel(val, addr);
}


/* memory allocation */
struct _mhead {
    unsigned int    mtype;
    unsigned int    msize;
    void *          maddr;
};

void *os_malloc(unsigned int size)
{
    void*           addr_start = NULL;
    void*           addr_data = NULL;
    struct _mhead *  mhead = NULL;
    unsigned int    memsize;
    unsigned char   mtype;
    unsigned long   kmalloc_max_size;
    unsigned int    alignment;
    
    if (size == 0) {
        return (0);
    }

    // 4k plus dma need align to 16 bytes
    alignment = sizeof(*mhead) <= 16 ? 16 : 32;
    memsize = alignment + size;

#ifdef KMALLOC_MAX_CACHE_SIZE
    kmalloc_max_size = KMALLOC_MAX_CACHE_SIZE;
#else
    kmalloc_max_size = PAGE_SIZE;
#endif

    if (memsize <= kmalloc_max_size) {
        addr_start = kmalloc(memsize, GFP_KERNEL);
        mtype = OS_LINUX_MEM_TYPE_KMALLOC;
    }

    if (addr_start == NULL) {
        memsize = PAGE_ALIGN(memsize);
        addr_start = vmalloc(memsize);
        mtype = OS_LINUX_MEM_TYPE_VMALLOC;
    }

    if (!addr_start) {
        return (0);
    }

    addr_data = addr_start + alignment;
    mhead = (struct _mhead *)(addr_data);
    mhead--;

    mhead->maddr = addr_start;
    mhead->msize = memsize;
    mhead->mtype = mtype;

    return addr_data;
}

static inline bool is_pow_of_two(unsigned long num)
{
    if (num < 1) return false;
    return (num & (num - 1)) == 0;
}

void *os_malloc_aligned(unsigned int size, unsigned int alignment)
{
    void*           addr_start = NULL;
    void*           addr_data = NULL;
    struct _mhead *  mhead = NULL;
    unsigned int    memsize;

    if (size == 0 || !is_pow_of_two(alignment) || alignment < 32) {
        return (0);
    }

    memsize = alignment + size;
    memsize = PAGE_ALIGN(memsize);
    addr_start = (struct _mhead *)vmalloc(memsize);

    if (!addr_start) {
        return (0);
    }

    addr_data = addr_start + alignment;
    mhead = (struct _mhead *)(addr_data);
    mhead--;

    mhead->maddr = addr_start;
    mhead->msize = memsize;
    mhead->mtype = OS_LINUX_MEM_TYPE_VMALLOC;

    return addr_data;
}

void *os_zalloc(unsigned int size)
{
    void * addr;

    addr = os_malloc(size);

    if (addr != NULL)
        memset(addr, 0, size);

    return addr;
}

void os_free(void *address)
{
    struct _mhead * hdr;

    hdr = (struct _mhead *)address;
    hdr--;

    if (hdr->mtype == OS_LINUX_MEM_TYPE_KMALLOC)
        kfree(hdr->maddr);
    else
        vfree(hdr->maddr);
}

size_t os_mem_get_actual_size(void *address)
{
    struct _mhead * hdr;

    hdr = (struct _mhead *)address;
    hdr--;

    if (hdr->mtype == OS_LINUX_MEM_TYPE_KMALLOC)
        return ksize(hdr->maddr) - (address - hdr->maddr);
    else
        return hdr->msize - (address - hdr->maddr);
}

int os_mem_get_mem_type(void *address)
{
    struct _mhead * hdr;

    hdr = (struct _mhead *)address;
    hdr--;

    return hdr->mtype;
}


struct _os_contig_dma_desc_t {
    dma_addr_t          phys_addr;
    void                *virt_addr;
    size_t              size;

    void                *parent;
};

os_contig_dma_desc_t os_contig_dma_alloc(size_t size, os_dma_par_t par)
{
    int nr_pages = PAGE_ALIGN(size) >> PAGE_SHIFT;
    void *parent = par;
    os_contig_dma_desc_t desc;

    desc = os_zalloc(sizeof(*desc));
    if (desc == NULL)
        return NULL;

    desc->size = nr_pages << PAGE_SHIFT;
    desc->virt_addr = dma_alloc_coherent(parent, desc->size, &desc->phys_addr, GFP_KERNEL);
    if (NULL == desc->virt_addr) {
        os_free(desc);
        return NULL;
    }

    desc->parent = parent;

    return desc;
}

void os_contig_dma_free(os_contig_dma_desc_t desc)
{
    if (desc == NULL)
        return;

    if (desc->virt_addr != NULL)
        dma_free_coherent(desc->parent, desc->size, desc->virt_addr, desc->phys_addr);

    os_free(desc);
}

void *os_contig_dma_get_virt(os_contig_dma_desc_t desc)
{
    return desc->virt_addr;
}

mw_physical_addr_t os_contig_dma_get_phys(os_contig_dma_desc_t desc)
{
    return desc->phys_addr;
}

size_t os_contig_dma_get_size(os_contig_dma_desc_t desc)
{
    return desc->size;
}


/* copy */
int os_copy_in(void *kaddr, const os_user_addr_t uaddr, size_t len)
{
    if (copy_from_user(kaddr, uaddr, len) == 0)
        return OS_RETURN_SUCCESS;

    return OS_RETURN_ERROR;
}

int os_copy_out(os_user_addr_t uaddr, const void *kaddr, size_t len)
{
    if (copy_to_user(uaddr, kaddr, len) == 0)
        return OS_RETURN_SUCCESS;

    return OS_RETURN_ERROR;
}


/* internal copy */
void *os_memset(void *dst, int ch, unsigned int size)
{
    return memset(dst, ch, (size_t)size);
}

void *os_memcpy(void *dst, const void *src, unsigned int size)
{
    return memcpy(dst, src, (size_t)size);
}

void *os_memmove(void *dst, const void *src, unsigned int size)
{
    return memmove(dst, src, (size_t)size);
}

int os_memcmp(const void *s1, const void *s2, unsigned int size)
{
    return memcmp(s1, s2, (size_t)size);
}

unsigned int os_strlcpy(char *dst, const char *src, unsigned int size)
{
    return strlcpy(dst, src, (size_t)size);
}

unsigned int os_strlcat(char *dst, const char *src, unsigned int size)
{
    return strlcat(dst, src, (size_t)size);
}


/* math */
#ifndef __LP64__
long long os_div64(long long dividend, long long divisor)
{
    return div64_s64(dividend, divisor);
}
#endif


/* pci config read/write */
int os_pci_read_config_dword(os_pci_dev_t dev, int addr, unsigned int *value)
{
    if (dev == NULL)
        return OS_RETURN_INVAL;

    return pci_read_config_dword(dev, addr, value);
}

int os_pci_write_config_dword(os_pci_dev_t dev, int addr, unsigned int value)
{
    if (dev == NULL)
        return OS_RETURN_INVAL;

    return pci_write_config_dword(dev, addr, value);
}


/* something */
void os_set_freezable()
{
    set_freezable();
}

bool os_try_to_freeze()
{
    return try_to_freeze();
}


/* zlib */
#include <linux/zlib.h>
/* Returns length of decompressed data. */
int os_zlib_uncompress(uint8_t *dst, size_t *dstlen, uint8_t *src, size_t srclen)
{
    z_stream stream;
    int err;

    if (dstlen == NULL || *dstlen <= 0)
        return OS_RETURN_INVAL;

    stream.workspace = vmalloc(zlib_inflate_workspacesize());
    if ( !stream.workspace ) {
        return OS_RETURN_NOMEM;
    }
    stream.next_in = NULL;
    stream.avail_in = 0;
    zlib_inflateInit(&stream);

    stream.next_in = src;
    stream.avail_in = srclen;

    stream.next_out = dst;
    stream.avail_out = *dstlen;

    err = zlib_inflateReset(&stream);
    if (err != Z_OK) {
        printk("zlib_inflateReset error %d\n", err);
        zlib_inflateEnd(&stream);
        zlib_inflateInit(&stream);
    }

    err = zlib_inflate(&stream, Z_FINISH);
    if (err != Z_STREAM_END)
        goto err;

    zlib_inflateEnd(&stream);
    vfree(stream.workspace);

    *dstlen = stream.total_out;

    return OS_RETURN_SUCCESS;

err:
    printk("Error %d while decompressing!\n", err);
    //printk("%p(%d)->%p(%d)\n", src, srclen, dst, dstlen);
    zlib_inflateEnd(&stream);
    vfree(stream.workspace);
    return OS_RETURN_ERROR;
}

